<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Stack\File;

// Import framework stuff
use Org\Mxchange\CoreFramework\Bootstrap\FrameworkBootstrap;
use Org\Mxchange\CoreFramework\Filesystem\Stack\FileStacker;
use Org\Mxchange\CoreFramework\Filesystem\File\BaseBinaryFile;
use Org\Mxchange\CoreFramework\Generic\UnsupportedOperationException;
use Org\Mxchange\CoreFramework\Stack\File\StackableFile;
use Org\Mxchange\CoreFramework\Utils\Crypto\CryptoUtils;
use Org\Mxchange\CoreFramework\Utils\Strings\StringUtils;

// Import SPL stuff
use \BadMethodCallException;
use \InvalidArgumentException;
use \SplFileInfo;

/**
 * A stack file class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2022 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
class StackFile extends BaseBinaryFile implements FileStacker {
	/**
	 * Protected constructor
	 *
	 * @return	void
	 */
	private function __construct () {
		// Call parent constructor
		parent::__construct(__CLASS__);
	}

	/**
	 * Creates an instance of this File class and prepares it for usage
	 *
	 * @param	$infoInstance	An instance of a SplFileInfo class
	 * @param	$stackInstance	An instance of a StackableFile class
	 * @return	$stackFileInstance	An instance of this File class
	 */
	public final static function createStackFile (SplFileInfo $infoInstance, StackableFile $stackInstance) {
		// Get a new instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: infoInstance[%s]=%s,stackInstance=%s - CALLED!', get_class($infoInstance), $infoInstance, $stackInstance->__toString()));
		$stackFileInstance = new StackFile();

		// Set stack instance here for callbacks
		$stackFileInstance->setStackInstance($stackInstance);

		// Init this abstract file
		$stackFileInstance->initFile($infoInstance);

		// Init counters and gaps array
		$stackFileInstance->initCountersGapsArray();

		// Return the prepared instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: stackFileInstance=%s - EXIT!', $stackFileInstance->__toString()));
		return $stackFileInstance;
	}

	/**
	 * Flushes the file header
	 *
	 * @return	void
	 * @throws	BadMethodCallException	If this->stackInstance is not properly set
	 */
	public function flushFileHeader () {
		// Validate call
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: CALLED!');
		if (!($this->getStackInstance() instanceof StackableFIle)) {
			// Index instance not set
			throw new BadMethodCallException('this->stackInstance[] is not properly set.');
		}

		// Call block instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: Calling this->indexInstance->flushFileHeader() ...');
		$this->getStackInstance()->flushFileHeader();

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: EXIT!');
	}

	/**
	 * Pre-allocates file (if enabled) with some space for later faster write access.
	 *
	 * @param	$type	Type of the file
	 * @return	void
	 * @throws	InvalidArgumentException	If a parameter is empty
	 * @throws	BadMethodCallException	If this->stackInstance is not properly set
	 */
	public function preAllocateFile (string $type) {
		// Is it enabled?
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: type=%s - CALLED!', $type));
		if (empty($type)) {
			// Empty type
			throw new InvalidArgumentException('Parameter "type" is empty');
		} elseif (!($this->getStackInstance() instanceof StackableFile) && !($this->getStackInstance() instanceof StackableFile)) {
			// Index instance not set
			throw new BadMethodCallException('this->stackInstance[] and this->pointerInstance are not properly set.');
		}

		// Message to user
		self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: Pre-allocating file ...');

		// Calculate minimum block length and get file size
		$minimumBlockLength = $this->getStackInstance()->calculateMinimumBlockLength();

		// Call protected method
		$this->preAllocateFileByTypeLength($type, $minimumBlockLength);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: EXIT!');
	}

	/**
	 * Checks wether the current entry is valid (not at the end of the file).
	 * This method will return true if an emptied (nulled) entry has been found.
	 *
	 * @return	$isValid	Whether the next entry is valid
	 * @throws	UnexpectedValueException	If some value is not expected
	 * @throws	BadMethodCallException	If this->stackInstance is not properly set
	 */
	public function isValid () {
		// Validate call
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: CALLED!');
		if (!($this->getStackInstance() instanceof StackableFile)) {
			// Index instance not set
			throw new BadMethodCallException('this->stackInstance[] is not properly set.');
		}

		// Get length from index
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: Calling this->stackInstance->calculateMinimumBlockLength() ...');
		$length = $this->getStackInstance()->calculateMinimumBlockLength();

		// Call protected method
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: Calling this->isValidByLength(%d) ...', $length));
		$isValid = $this->isValidByLength($length);

		// Return result
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: isValid=%d - EXIT!', intval($isValid)));
		return $isValid;
	}

	/**
	 * Reads the file header
	 *
	 * @return	void
	 * @throws	LogicException	If both instances are not set
	 */
	public function readFileHeader () {
		// Call stacke instance
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: Calling this->stackInstance->readStackHeader() - CALLED!');
		$this->getStackInstance()->readStackHeader();

		// Trace message
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: EXIT!');
	}

	/**
	 * Reads next "block" of bytes into $currentBlock field. THis method loads
	 * the whole file into memory when the file is just freshly initialized
	 * (only zeros in it).
	 *
	 * @return	void
	 */
	protected function readNextBlock () {
		// First calculate minimum block length
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: this->seekPosition=%d - CALLED!', $this->getSeekPosition()));
		$length = $this->getStackInstance()->calculateMinimumBlockLength();

		// Call protected method
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: Calling parent::readNextBlockByLength(%d) ...', $length));
		parent::readNextBlockByLength($length);

		// Trace message
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: EXIT!');
	}

	/**
	 * Writes given value to the file and returns a hash and gap position for it
	 *
	 * @param	$stackName	Group identifier
	 * @param	$value		Value to be added to the stack
	 * @return	$data		Hash and gap position
	 * @throws	InvalidArgumentException	If a parameter is not valid
	 */
	public function writeValueToFile (string $stackName, $value) {
		// Validate parameter
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: stackName=%s,value[]=%s - CALLED!', $stackName, gettype($value)));
		if (empty($stackName)) {
			// Throw IAE
			throw new InvalidArgumentException('Parameter "stackName" is empty');
		} elseif (is_object($value) || is_resource($value)) {
			// Not wanted here
			throw new InvalidArgumentException(sprintf('value[]=%s is not stackable in files', gettype($value)));
		}

		// Encode/convert the value into a "binary format"
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: Calling StringUtils::encodeData(value[]=%s) ...', gettype($value)));
		$encoded = StringUtils::encodeData($value);

		// Get a strong hash for the "encoded" data
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: Calling CryptoUtils::hash(%s) ...', $encoded));
		$hash = CryptoUtils::hash($encoded);

		// Then write it to the next free gap
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: Calling this->stackInstance->writeDataToFreeGap(%s,%s,%s) ...', $stackName, $hash, $encoded));
		$data = $this->getStackInstance()->writeDataToFreeGap($stackName, $hash, $encoded);

		// Return info
		/* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput(sprintf('STACK-FILE: data[]=%s - EXIT!', gettype($data)));
		return $data;
	}

	/**
	 * Writes given raw data to the file and returns a gap position and length
	 *
	 * @param	$stackName	Group identifier
	 * @param	$hash		Hash from encoded value
	 * @param	$encoded	Encoded value to be written to the file
	 * @return	$data		Gap position and length of the raw data
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public function writeDataToFreeGap (string $stackName, string $hash, string $encoded) {
		self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('STACK-FILE: stackName=' . $stackName . ',hash=' . $hash . ',encoded()=' . strlen($encoded));
		throw new UnsupportedOperationException([$this, __FUNCTION__], self::EXCEPTION_UNSPPORTED_OPERATION);
	}

}
